
class XmlConstants:
    """
    A `no method` class that just hosts the strings that represent the UML shapes XML strings
    """
    ATTRIBUTE_VERSION: str = 'version'

    ATTRIBUTE_DOCUMENT_TYPE: str = 'documentType'
    ATTRIBUTE_TITLE:         str = 'title'

    ATTRIBUTE_CODE_PATH: str = 'codePath'

    ATTRIBUTE_SCROLL_POSITION_X: str = 'scrollPositionX'
    ATTRIBUTE_SCROLL_POSITION_Y: str = 'scrollPositionY'
    ATTRIBUTE_PIXELS_PER_UNIT_X: str = 'pixelsPerUnitX'
    ATTRIBUTE_PIXELS_PER_UNIT_Y: str = 'pixelsPerUnitY'

    ATTRIBUTE_ID:     str = 'id'

    ATTRIBUTE_WIDTH:  str = 'width'
    ATTRIBUTE_HEIGHT: str = 'height'
    ATTRIBUTE_X:      str = 'x'
    ATTRIBUTE_Y:      str = 'y'
    ATTRIBUTE_NAME:   str = 'name'

    ATTRIBUTE_DISPLAY_STEREOTYPE:     str = 'displayStereotype'
    ATTRIBUTE_DISPLAY_METHODS:        str = 'displayMethods'
    ATTRIBUTE_DISPLAY_FIELDS:         str = 'displayFields'
    ATTRIBUTE_DISPLAY_PARAMETERS:     str = 'displayParameters'
    ATTRIBUTE_DISPLAY_CONSTRUCTOR:    str = 'displayConstructor'
    ATTRIBUTE_DISPLAY_DUNDER_METHODS: str = 'displayDunderMethods'

    ATTRIBUTE_STEREOTYPE:    str = 'stereotype'
    ATTRIBUTE_FILENAME:      str = 'fileName'
    ATTRIBUTE_CONTENT:       str = 'content'
    ATTRIBUTE_DESCRIPTION:   str = 'description'
    ATTRIBUTE_VISIBILITY:    str = 'visibility'
    ATTRIBUTE_MESSAGE:       str = 'message'
    ATTRIBUTE_DEFAULT_VALUE: str = 'defaultValue'
    ATTRIBUTE_METHOD_RETURN_TYPE: str = 'returnType'

    ATTRIBUTE_SOURCE_CARDINALITY_VALUE:      str = 'sourceCardinalityValue'
    ATTRIBUTE_DESTINATION_CARDINALITY_VALUE: str = 'destinationCardinalityValue'

    ATTRIBUTE_DELTA_X: str = 'deltaX'
    ATTRIBUTE_DELTA_Y: str = 'deltaY'

    ATTRIBUTE_LINK_FROM_X: str = 'fromX'
    ATTRIBUTE_LINK_FROM_Y: str = 'fromY'

    ATTRIBUTE_LINK_TO_X: str = 'toX'
    ATTRIBUTE_LINK_TO_Y: str = 'toY'

    ATTRIBUTE_SPLINE:    str = 'spline'
    ATTRIBUTE_LINK_TYPE: str = 'type'       # TODO:  Should be linkType

    ATTRIBUTE_SOURCE_ID:                 str = 'sourceId'
    ATTRIBUTE_DESTINATION_ID:            str = 'destinationId'
    ATTRIBUTE_BIDIRECTIONAL:             str = 'bidirectional'
    ATTRIBUTE_SD_MESSAGE_SOURCE_ID:      str = 'sourceId'
    ATTRIBUTE_SD_MESSAGE_DESTINATION_ID: str = 'destinationId'

    ATTRIBUTE_INSTANCE_NAME:    str = 'instanceName'
    ATTRIBUTE_LIFE_LINE_LENGTH: str = 'lifeLineLength'
    ATTRIBUTE_SOURCE_TIME:      str = 'sourceTime'
    ATTRIBUTE_DESTINATION_TIME: str = 'destinationTime'

    ATTRIBUTE_IMPLEMENTING_CLASS_NAME:   str = 'implementingClassName'
    ATTRIBUTE_LINE_CENTUM:               str = 'lineCentum'
    ATTRIBUTE_ATTACHMENT_SIDE:           str = 'attachmentSide'
    ATTRIBUTE_ATTACHED_TO_ID:            str = 'attachedToId'

    ATTRIBUTE_FIELD_TYPE:     str = 'fieldType'
    ATTRIBUTE_PARAMETER_TYPE: str = 'parameterType'

    ELEMENT_UML_PROJECT:  str = 'UmlProject'
    ELEMENT_UML_DIAGRAM:  str = 'UMLDiagram'
    ELEMENT_UML_CLASS:    str = 'UmlClass'
    ELEMENT_UML_USE_CASE: str = 'UmlUseCase'
    ELEMENT_UML_ACTOR:    str = 'UmlActor'
    ELEMENT_UML_NOTE:     str = 'UmlNote'
    ELEMENT_UML_TEXT:     str = 'UmlText'
    ELEMENT_UML_LINK:     str = 'UmlLink'
    ELEMENT_LOLLIPOP:     str = 'UmlLollipopInterface'

    ELEMENT_ASSOCIATION_LABEL:             str = 'AssociationName'
    ELEMENT_ASSOCIATION_SOURCE_LABEL:      str = 'SourceCardinality'
    ELEMENT_ASSOCIATION_DESTINATION_LABEL: str = 'DestinationCardinality'

    ELEMENT_MODEL_CLASS:       str = 'PyutClass'
    ELEMENT_MODEL_TEXT:        str = 'PyutText'
    ELEMENT_MODEL_NOTE:        str = 'PyutNote'
    ELEMENT_MODEL_ACTOR:       str = 'PyutActor'
    ELEMENT_MODEL_USE_CASE:    str = 'PyutUseCase'
    ELEMENT_MODEL_LINK:        str = 'PyutLink'
    ELEMENT_MODEL_INTERFACE:   str = 'PyutInterface'
    ELEMENT_MODEL_IMPLEMENTOR: str = 'Implementor'

    ELEMENT_MODEL_IMPLEMENTING_CLASS_NAME: str = 'implementingClassName'

    ELEMENT_MODEL_METHOD:      str = 'PyutMethod'
    ELEMENT_MODEL_PARAMETER:   str = 'PyutParameter'
    ELEMENT_MODEL_FIELD:       str = 'PyutField'
    ELEMENT_MODEL_MODIFIER:    str = 'Modifier'
    ELEMENT_MODEL_SOURCE_CODE: str = 'SourceCode'
    ELEMENT_MODEL_CODE:        str = 'Code'

    ELEMENT_MODEL_SD_INSTANCE:   str = 'PyutSDInstance'
    ELEMENT_MODEL_SD_MESSAGE:    str = 'PyutSDMessage'

    ELEMENT_MODEL_LINE_CONTROL_POINT: str = 'LineControlPoint'




from typing import cast

from logging import Logger
from logging import getLogger

from wx import Point


from xml.etree.ElementTree import Element
from xml.etree.ElementTree import SubElement

from umlshapes.types.DeltaXY import DeltaXY
from umlshapes.links.UmlLink import UmlLink
from umlshapes.links.UmlAssociation import UmlAssociation
from umlshapes.links.UmlAssociationLabel import UmlAssociationLabel

from umlshapes.links.eventhandlers.UmlLinkEventHandler import LineControlPoints

from umlshapes.types.Common import EndPoints
from umlshapes.types.UmlPosition import UmlPosition

from umlio.IOTypes import ElementAttributes
from umlio.IOTypes import UmlLinks

from umlio.serializer.PyutToXml import PyutToXml
from umlio.serializer.BaseUmlToXml import BaseUmlToXml

from umlio.XMLConstants import XmlConstants


class UmlLinksToXml(BaseUmlToXml):
    def __init__(self):
        super().__init__()
        self.logger: Logger = getLogger(__name__)

        self._pyutToXml: PyutToXml = PyutToXml()

    def serialize(self, documentTop: Element, umlLinks: UmlLinks) -> Element:

        for umlLink in umlLinks:
            self._umlLinkToXml(documentElement=documentTop, umlLink=umlLink)

        return documentTop

    def _umlLinkToXml(self, documentElement: Element, umlLink: UmlLink) -> Element:

        attributes:        ElementAttributes = self._umlLinkAttributes(umlLink=umlLink)
        oglLinkSubElement: Element           = SubElement(documentElement, XmlConstants.ELEMENT_UML_LINK, attrib=attributes)

        if isinstance(umlLink, UmlAssociation):

            associationName: UmlAssociationLabel = umlLink.associationName
            src:             UmlAssociationLabel = umlLink.sourceCardinality
            dst:             UmlAssociationLabel = umlLink.destinationCardinality
            associationLabels = {
                XmlConstants.ELEMENT_ASSOCIATION_LABEL:      associationName,
                XmlConstants.ELEMENT_ASSOCIATION_SOURCE_LABEL:      src,
                XmlConstants.ELEMENT_ASSOCIATION_DESTINATION_LABEL: dst
            }
            for eltName in associationLabels:
                umlAssociationLabel: UmlAssociationLabel = associationLabels[eltName]

                linkDelta: DeltaXY = umlAssociationLabel.linkDelta

                labelAttributes: ElementAttributes = ElementAttributes({
                    XmlConstants.ATTRIBUTE_DELTA_X: str(linkDelta.deltaX),
                    XmlConstants.ATTRIBUTE_DELTA_Y: str(linkDelta.deltaY),
                })
                # noinspection PyUnusedLocal
                labelElement: Element = SubElement(oglLinkSubElement, eltName, attrib=labelAttributes)

        lineControlPoints: LineControlPoints = umlLink.GetLineControlPoints()
        realControlPoints: LineControlPoints = self._removeEndPoints(umlLink=umlLink, lineControlPoints=lineControlPoints)
        for pt in realControlPoints:
            wxPoint: Point = cast(Point, pt)
            controlPointAttributes: ElementAttributes = ElementAttributes({
                XmlConstants.ATTRIBUTE_X: str(wxPoint.x),
                XmlConstants.ATTRIBUTE_Y: str(wxPoint.y),
            })
            SubElement(oglLinkSubElement, XmlConstants.ELEMENT_MODEL_LINE_CONTROL_POINT, attrib=controlPointAttributes)

        self._pyutToXml.pyutLinkToXml(pyutLink=umlLink.pyutLink, umlLinkElement=oglLinkSubElement)

        return oglLinkSubElement

    def _umlLinkAttributes(self, umlLink: UmlLink) -> ElementAttributes:

        # srcX, srcY   = umlLink.sourceAnchor.model.GetPosition()
        # destX, destY = umlLink.destinationAnchor.model.GetPosition()

        umlLinkId:    str         = str(umlLink.id)
        endPoints:    EndPoints   = umlLink.endPoints
        fromPosition: UmlPosition = endPoints.fromPosition
        toPosition:   UmlPosition = endPoints.toPosition

        attributes: ElementAttributes = ElementAttributes({
            XmlConstants.ATTRIBUTE_ID:          umlLinkId,
            XmlConstants.ATTRIBUTE_LINK_FROM_X: str(fromPosition.x),
            XmlConstants.ATTRIBUTE_LINK_FROM_Y: str(fromPosition.y),
            XmlConstants.ATTRIBUTE_LINK_TO_X:   str(toPosition.x),
            XmlConstants.ATTRIBUTE_LINK_TO_Y:   str(toPosition.y),
            XmlConstants.ATTRIBUTE_SPLINE:      str(umlLink.spline)   # piecewise polynomial function
        })

        return attributes

    def _removeEndPoints(self, umlLink: 'UmlLink', lineControlPoints: LineControlPoints) -> LineControlPoints:
        """
        Do not consider the end points

        Args:
            umlLink:
            lineControlPoints:

        Returns:  The control points less the 2 end points
        """

        realControlPoints: LineControlPoints = LineControlPoints(lineControlPoints[:])

        x1, y1, x2, y2 = umlLink.FindLineEndPoints()

        pt1: Point = Point(x1, y1)
        pt2: Point = Point(x2, y2)

        realControlPoints.remove(pt1)
        realControlPoints.remove(pt2)

        return realControlPoints




from logging import Logger
from logging import getLogger
from typing import cast

from codeallybasic.SecureConversions import SecureConversions
from pyutmodelv2.PyutInterface import PyutInterface
from umlshapes.links.UmlLollipopInterface import UmlLollipopInterface
from umlshapes.shapes.UmlClass import UmlClass
from umlshapes.types.Common import AttachmentSide
from untangle import Element

from umlshapes.ShapeTypes import LinkableUmlShapes

from umlio.IOTypes import Elements
from umlio.IOTypes import UmlLollipopInterfaces
from umlio.IOTypes import umlLollipopInterfacesFactory

from umlio.XMLConstants import XmlConstants

from umlio.deserializer.XmlToPyut import XmlToPyut


class XmlLollipopsToUmlLollipops:
    def __init__(self):

        self.logger: Logger = getLogger(__name__)

        self._xmlToPyut: XmlToPyut = XmlToPyut()

    def deserialize(self, umlDiagramElement: Element, linkableUmlShapes: LinkableUmlShapes) -> UmlLollipopInterfaces:

        umlLollipops:     UmlLollipopInterfaces = umlLollipopInterfacesFactory()
        lollipopElements: Elements              = cast(Elements, umlDiagramElement.get_elements(XmlConstants.ELEMENT_LOLLIPOP))

        for lollipopElement in lollipopElements:
            self.logger.info(f'{lollipopElement}')
            umlLollipopInterface: UmlLollipopInterface = self._getLollipop(lollipopElement=lollipopElement, linkableUmlShapes=linkableUmlShapes)

            umlLollipops.append(umlLollipopInterface)

        return umlLollipops

    def _getLollipop(self, lollipopElement: Element, linkableUmlShapes: LinkableUmlShapes) -> UmlLollipopInterface:
        """
        <UmlLollipopInterface lineCentum="0.1" attachmentSide="Right" attachedToId="valley.darkness.implementor">

        Args:
            lollipopElement:
            linkableUmlShapes:

        Returns:   A UML Lollipop interface class
        """
        pyutInterface: PyutInterface = self._xmlToPyut.interfaceToPyutInterface(lollipopElement)

        umlLollipopInterface: UmlLollipopInterface = UmlLollipopInterface(pyutInterface=pyutInterface)

        attachmentSideStr: str      = lollipopElement[XmlConstants.ATTRIBUTE_ATTACHMENT_SIDE]
        attachedToId:      str      = lollipopElement[XmlConstants.ATTRIBUTE_ATTACHED_TO_ID]
        attachedTo:        UmlClass = self._findAttachedToClass(attachedToId=attachedToId, linkableUmlShapes=linkableUmlShapes)

        umlLollipopInterface.lineCentum     = SecureConversions.secureFloat(lollipopElement[XmlConstants.ATTRIBUTE_LINE_CENTUM])
        umlLollipopInterface.attachmentSide = AttachmentSide.toEnum(attachmentSideStr)

        umlLollipopInterface.attachedTo = attachedTo

        return umlLollipopInterface

    def _findAttachedToClass(self, attachedToId: str, linkableUmlShapes: LinkableUmlShapes) -> UmlClass:
        """
        This method is necessary because the linkable shapes dictionary is indexed by the PyutClass ID
        However, the pointer to the attached class is a UmlClass id.  All this came about because
        lollipops reused the PyutInterface class

        Args:
            attachedToId:
            linkableUmlShapes:

        Returns:  The appropriate UmlClass
        """
        foundClass: UmlClass = cast(UmlClass, None)
        for uc in linkableUmlShapes.values():
            umlClass: UmlClass = cast(UmlClass, uc)
            if umlClass.id == attachedToId:
                foundClass = umlClass
                break

        assert foundClass is not None, 'Developer error, missing class in linkableUmlShapes'
        return foundClass




from logging import Logger
from logging import getLogger

from xml.etree.ElementTree import Element
from xml.etree.ElementTree import SubElement

from umlshapes.shapes.UmlClass import UmlClass

from umlio.IOTypes import UmlClasses

from umlio.serializer.BaseUmlToXml import BaseUmlToXml
from umlio.serializer.PyutToXml import PyutToXml
from umlio.XMLConstants import XmlConstants


class UmlClassToXml(BaseUmlToXml):
    def __init__(self):

        super().__init__()

        self.logger:     Logger    = getLogger(__name__)
        self._pyutToXml: PyutToXml = PyutToXml()

    def serialize(self, documentTop: Element, umlClasses: UmlClasses) -> Element:

        for umlClass in umlClasses:
            self._umlClassToXml(documentTop=documentTop, umlClass=umlClass)

        return documentTop

    def _umlClassToXml(self, documentTop: Element, umlClass: UmlClass) -> Element:
        """
        Exports an UmlClass to a minidom Element.

        Args:
            documentTop: The document to append to
            umlClass:    UML Class to serialize

        Returns:
            The newly created `UmlClass` Element
        """
        attributes = self._umlBaseAttributes(umlShape=umlClass)
        umlClassSubElement: Element = SubElement(documentTop, XmlConstants.ELEMENT_UML_CLASS, attrib=attributes)

        self._pyutToXml.pyutClassToXml(umlClassElement=umlClassSubElement, pyutClass=umlClass.pyutClass)

        return umlClassSubElement




from umlshapes.types.Common import UmlShape
from umlshapes.types.UmlDimensions import UmlDimensions
from umlshapes.types.UmlPosition import UmlPosition

from umlio.IOTypes import ElementAttributes
from umlio.XMLConstants import XmlConstants


class BaseUmlToXml:
    def __init__(self):
        pass

    def _umlBaseAttributes(self, umlShape: UmlShape) -> ElementAttributes:
        """
        Create the common OglObject attributes

        Args:
            umlShape:  OGL Object

        Returns:
            The updated originalElement
        """
        umlClassId: str           = str(umlShape.id)
        size:       UmlDimensions = umlShape.size
        position:   UmlPosition   = umlShape.position

        w: int = size.width
        h: int = size.height
        x: int = position.x
        y: int = position.y

        attributes: ElementAttributes = ElementAttributes(
            {
                XmlConstants.ATTRIBUTE_ID:     umlClassId,
                XmlConstants.ATTRIBUTE_WIDTH:  str(w),
                XmlConstants.ATTRIBUTE_HEIGHT: str(h),
                XmlConstants.ATTRIBUTE_X:      str(x),
                XmlConstants.ATTRIBUTE_Y:      str(y),
            }
        )

        return attributes




from typing import cast

from logging import Logger
from logging import getLogger

from untangle import Element

from pyutmodelv2.PyutClass import PyutClass

from umlshapes.shapes.UmlClass import UmlClass

from umlio.IOTypes import Elements
from umlio.IOTypes import GraphicInformation
from umlio.IOTypes import UmlClasses
from umlio.IOTypes import umlClassesFactory

from umlio.XMLConstants import XmlConstants

from umlio.deserializer.XmlToPyut import XmlToPyut


class XmlClassesToUmlClasses:
    def __init__(self):
        self.logger: Logger = getLogger(__name__)

        self._xmlToPyut: XmlToPyut = XmlToPyut()

    def deserialize(self, umlDiagramElement: Element) -> UmlClasses:
        """

        Args:
            umlDiagramElement:  The Element document

        Returns:  deserialized UmlNote objects if any exist, else an empty list
        """
        umlClasses:    UmlClasses = umlClassesFactory()
        classElements: Elements   = cast(Elements, umlDiagramElement.get_elements(XmlConstants.ELEMENT_UML_CLASS))

        for classElement in classElements:
            self.logger.debug(f'{classElement}')

            graphicInformation: GraphicInformation = GraphicInformation.toGraphicInfo(graphicElement=classElement)
            pyutClass:          PyutClass          = self._xmlToPyut.classToPyutClass(umlClassElement=classElement)
            umlClass:           UmlClass           = UmlClass(pyutClass=pyutClass)

            umlClass.id       = graphicInformation.id
            umlClass.size     = graphicInformation.size
            umlClass.position = graphicInformation.position

            umlClasses.append(umlClass)

        return umlClasses




from typing import cast

from logging import Logger
from logging import getLogger

from pyutmodelv2.PyutActor import PyutActor
from umlshapes.shapes.UmlActor import UmlActor
from untangle import Element

from umlio.IOTypes import Elements
from umlio.IOTypes import GraphicInformation
from umlio.IOTypes import UmlActors
from umlio.IOTypes import umlActorsFactory

from umlio.XMLConstants import XmlConstants

from umlio.deserializer.XmlToPyut import XmlToPyut


class XmlActorsToUmlActors:
    def __init__(self):
        self.logger: Logger = getLogger(__name__)

        self._xmlToPyut: XmlToPyut = XmlToPyut()

    def deserialize(self, umlDiagramElement: Element) -> UmlActors:
        """

        Args:
            umlDiagramElement:  The Element document

        Returns:  deserialized UmlNote objects if any exist, else an empty list
        """
        umlActors:     UmlActors = umlActorsFactory()
        actorElements: Elements = cast(Elements, umlDiagramElement.get_elements(XmlConstants.ELEMENT_UML_ACTOR))

        for actorElement in actorElements:
            self.logger.debug(f'{actorElement}')

            graphicInformation: GraphicInformation = GraphicInformation.toGraphicInfo(graphicElement=actorElement)
            pyutActor:          PyutActor           = self._xmlToPyut.actorToPyutActor(umlActorElement=actorElement)
            umlActor:           UmlActor            = UmlActor(pyutActor=pyutActor)

            umlActor.id       = graphicInformation.id
            umlActor.size     = graphicInformation.size
            umlActor.position = graphicInformation.position

            umlActors.append(umlActor)

        return umlActors




from logging import Logger
from logging import getLogger
from typing import cast

from xml.etree.ElementTree import Element
from xml.etree.ElementTree import SubElement

from umlshapes.shapes.UmlActor import UmlActor
from umlshapes.shapes.UmlUseCase import UmlUseCase

from umlio.IOTypes import UmlActors
from umlio.IOTypes import UmlUseCases

from umlio.serializer.BaseUmlToXml import BaseUmlToXml
from umlio.serializer.PyutToXml import PyutToXml
from umlio.XMLConstants import XmlConstants


class UmlUseCasesToXml(BaseUmlToXml):
    """
    """
    def __init__(self):
        super().__init__()
        self.logger: Logger = getLogger(__name__)

        self._pyutToXml: PyutToXml = PyutToXml()

    def serialize(self, documentTop: Element, umlUseCases: UmlUseCases, umlActors: UmlActors) -> Element:

        for actor in umlActors:
            umlActor:        UmlActor = cast(UmlActor, actor)
            umlActorElement: Element  = self._umlActorToXml(documentTop=documentTop, umlActor=umlActor)
            self._pyutToXml.pyutActorToXml(pyutActor=umlActor.pyutActor, umlActorElement=umlActorElement)

        for useCase in umlUseCases:
            umlUseCase:        UmlUseCase = cast(UmlUseCase, useCase)
            umlUseCaseElement: Element    = self._umlUseCaseToXml(documentTop=documentTop, umlUseCase=umlUseCase)
            self._pyutToXml.pyutUseCaseToXml(pyutUseCase=umlUseCase.pyutUseCase, umlUseCaseElement=umlUseCaseElement)

        return documentTop

    def _umlUseCaseToXml(self, documentTop: Element, umlUseCase: UmlUseCase) -> Element:

        attributes = self._umlBaseAttributes(umlShape=umlUseCase)
        umlTextSubElement: Element = SubElement(documentTop, XmlConstants.ELEMENT_UML_USE_CASE, attrib=attributes)

        return umlTextSubElement

    def _umlActorToXml(self, documentTop: Element, umlActor: UmlActor) -> Element:

        attributes = self._umlBaseAttributes(umlShape=umlActor)
        umlActorSubElement: Element = SubElement(documentTop, XmlConstants.ELEMENT_UML_ACTOR, attrib=attributes)

        return umlActorSubElement




from typing import cast

from dataclasses import dataclass

from logging import Logger
from logging import getLogger

from os import linesep as osLineSep

from pyutmodelv2.PyutLink import LinkDestination
from pyutmodelv2.PyutLink import LinkSource
from untangle import Element

from codeallybasic.Common import XML_END_OF_LINE_MARKER
from codeallybasic.SecureConversions import SecureConversions

from pyutmodelv2.PyutField import PyutField
from pyutmodelv2.PyutField import PyutFields

from pyutmodelv2.PyutObject import PyutObject
from pyutmodelv2.PyutUseCase import PyutUseCase
from pyutmodelv2.PyutActor import PyutActor
from pyutmodelv2.PyutInterface import PyutInterface
from pyutmodelv2.PyutLink import PyutLink
from pyutmodelv2.PyutMethod import PyutMethods
from pyutmodelv2.PyutMethod import PyutParameters
from pyutmodelv2.PyutMethod import SourceCode
from pyutmodelv2.PyutParameter import PyutParameter
from pyutmodelv2.PyutMethod import PyutMethod
from pyutmodelv2.PyutMethod import PyutModifiers
from pyutmodelv2.PyutModifier import PyutModifier
from pyutmodelv2.PyutType import PyutType
from pyutmodelv2.PyutClass import PyutClass
from pyutmodelv2.PyutNote import PyutNote
from pyutmodelv2.PyutText import PyutText

from pyutmodelv2.PyutSDInstance import PyutSDInstance
from pyutmodelv2.PyutSDMessage import PyutSDMessage

from pyutmodelv2.enumerations.PyutStereotype import PyutStereotype
from pyutmodelv2.enumerations.PyutVisibility import PyutVisibility
from pyutmodelv2.enumerations.PyutDisplayParameters import PyutDisplayParameters
from pyutmodelv2.enumerations.PyutLinkType import PyutLinkType
from pyutmodelv2.enumerations.PyutDisplayMethods import PyutDisplayMethods

from umlio.IOTypes import Elements

from umlio.XMLConstants import XmlConstants


@dataclass
class ConvolutedPyutSDMessageInformation:
    """
    This class is necessary because I do not want to mix Ogl and pyutmodel code;  Unfortunately,
    the IDs of the PyutSDInstance are buried and require a lookup

    """
    pyutSDMessage: PyutSDMessage = cast(PyutSDMessage, None)
    sourceId:      int           = -1
    destinationId: int           = -1


class XmlToPyut:
    """
    Converts pyutmodel Version 11 XML to Pyut Objects
    """
    NOTE_NAME:   str = 'Note'
    noteCounter: int = 0

    def __init__(self):

        self.logger: Logger = getLogger(__name__)

    def classToPyutClass(self, umlClassElement: Element) -> PyutClass:
        classElement: Element = umlClassElement.PyutClass

        pyutClass: PyutClass = PyutClass()

        pyutClass = cast(PyutClass, self._addPyutObjectAttributes(pyutElement=classElement, pyutObject=pyutClass))

        displayStr:              str                   = classElement[XmlConstants.ATTRIBUTE_DISPLAY_PARAMETERS]
        displayParameters:       PyutDisplayParameters = PyutDisplayParameters(displayStr)
        displayConstructorStr:   str                   = classElement[XmlConstants.ATTRIBUTE_DISPLAY_CONSTRUCTOR]
        displayDunderMethodsStr: str                   = classElement[XmlConstants.ATTRIBUTE_DISPLAY_DUNDER_METHODS]

        displayConstructor:   PyutDisplayMethods = self._securePyutDisplayMethods(displayStr=displayConstructorStr)
        displayDunderMethods: PyutDisplayMethods = self._securePyutDisplayMethods(displayStr=displayDunderMethodsStr)

        showStereotype:     bool = bool(classElement[XmlConstants.ATTRIBUTE_DISPLAY_STEREOTYPE])
        showFields:         bool = bool(classElement[XmlConstants.ATTRIBUTE_DISPLAY_FIELDS])
        showMethods:        bool = bool(classElement[XmlConstants.ATTRIBUTE_DISPLAY_METHODS])
        stereotypeStr:      str  = classElement[XmlConstants.ATTRIBUTE_STEREOTYPE]
        fileName:           str  = classElement[XmlConstants.ATTRIBUTE_FILENAME]

        pyutClass.displayParameters    = displayParameters
        pyutClass.displayConstructor   = displayConstructor
        pyutClass.displayDunderMethods = displayDunderMethods

        pyutClass.displayStereoType = showStereotype
        pyutClass.showFields        = showFields
        pyutClass.showMethods       = showMethods

        pyutClass.description = classElement['description']
        pyutClass.stereotype  = PyutStereotype.toEnum(stereotypeStr)
        pyutClass.fileName    = fileName

        pyutClass.methods = self._methodToPyutMethods(classElement=classElement)
        pyutClass.fields  = self._fieldToPyutFields(classElement=classElement)

        return pyutClass

    def textToPyutText(self, umlTextElement: Element) -> PyutText:
        """
        Parses the Text elements
        Args:
            umlTextElement:   Of the form:   <UmlText id="xx.xx.xx"/>

        Returns: A PyutText Object
        """
        pyutTextElement: Element  = umlTextElement.PyutText
        pyutText:    PyutText = PyutText()

        pyutText.id  = int(pyutTextElement[XmlConstants.ATTRIBUTE_ID])

        rawContent:   str = pyutTextElement['content']
        cleanContent: str = rawContent.replace(XML_END_OF_LINE_MARKER, osLineSep)
        pyutText.content = cleanContent

        return pyutText

    def noteToPyutNote(self, umlNoteElement: Element) -> PyutNote:
        """
        Parse Note element

        Args:
            umlNoteElement: of the form:  <UmlNote id="xx.xx.xx"/>

        Returns: A PyutNote Object
        """
        pyutNoteElement = umlNoteElement.PyutNote

        pyutNote: PyutNote = PyutNote()

        # fix line feeds
        pyutNote = cast(PyutNote, self._addPyutObjectAttributes(pyutElement=pyutNoteElement, pyutObject=pyutNote))

        rawContent:   str = pyutNoteElement[XmlConstants.ATTRIBUTE_CONTENT]
        cleanContent: str = rawContent.replace(XML_END_OF_LINE_MARKER, osLineSep)
        pyutNote.content = cleanContent

        return pyutNote

    def interfaceToPyutInterface(self, umlInterfaceElement: Element) -> PyutInterface:

        pyutInterfaceElement: Element = umlInterfaceElement.PyutInterface

        interfaceId: int = int(pyutInterfaceElement['id'])
        name:        str = pyutInterfaceElement[XmlConstants.ATTRIBUTE_NAME]
        description: str = pyutInterfaceElement[XmlConstants.ATTRIBUTE_DESCRIPTION]

        pyutInterface: PyutInterface = PyutInterface(name=name)
        pyutInterface.id          = interfaceId
        pyutInterface.description = description

        implementors: Elements = cast(Elements, pyutInterfaceElement.get_elements(XmlConstants.ELEMENT_MODEL_IMPLEMENTOR))
        for implementor in implementors:
            pyutInterface.addImplementor(implementor[XmlConstants.ELEMENT_MODEL_IMPLEMENTING_CLASS_NAME])

        pyutInterface.methods = self._interfaceMethodsToPyutMethods(interface=pyutInterfaceElement)

        return pyutInterface

    def actorToPyutActor(self, umlActorElement: Element) -> PyutActor:
        """

        Args:
            umlActorElement:   untangle Element in the above format

        Returns:   PyutActor
        """
        pyutActorElement: Element   = umlActorElement.PyutActor

        pyutActor: PyutActor = PyutActor()

        pyutActor = cast(PyutActor, self._addPyutObjectAttributes(pyutElement=pyutActorElement, pyutObject=pyutActor))

        return pyutActor

    def useCaseToPyutUseCase(self, umlUseCaseElement: Element) -> PyutUseCase:
        """

        Args:
            umlUseCaseElement:  An `untangle` Element in the above format

        Returns:  PyutUseCase
        """
        useCaseElement: Element     = umlUseCaseElement.PyutUseCase

        pyutUseCase:    PyutUseCase = PyutUseCase()

        pyutUseCase = cast(PyutUseCase, self._addPyutObjectAttributes(pyutElement=useCaseElement, pyutObject=pyutUseCase))

        return pyutUseCase

    def linkToPyutLink(self, singleLink: Element, source: LinkSource, destination: LinkDestination) -> PyutLink:

        linkTypeStr:     str          = singleLink[XmlConstants.ATTRIBUTE_LINK_TYPE]

        linkType:        PyutLinkType = PyutLinkType.toEnum(linkTypeStr)
        cardSrc:         str          = singleLink[XmlConstants.ATTRIBUTE_SOURCE_CARDINALITY_VALUE]
        cardDest:        str          = singleLink[XmlConstants.ATTRIBUTE_DESTINATION_CARDINALITY_VALUE]
        bidir:           bool         = SecureConversions.secureBoolean(singleLink[XmlConstants.ATTRIBUTE_BIDIRECTIONAL])
        linkDescription: str          = singleLink['name']

        pyutLink: PyutLink = PyutLink(name=linkDescription,
                                      linkType=linkType,
                                      cardinalitySource=cardSrc, cardinalityDestination=cardDest,
                                      bidirectional=bidir,
                                      source=source,
                                      destination=destination)

        return pyutLink

    def sdInstanceToPyutSDInstance(self, oglSDInstanceElement: Element) -> PyutSDInstance:

        instanceElement: Element = oglSDInstanceElement.PyutSDInstance
        pyutSDInstance:  PyutSDInstance = PyutSDInstance()

        pyutSDInstance.id                     = int(instanceElement[XmlConstants.ATTRIBUTE_ID])
        pyutSDInstance.instanceName           = instanceElement[XmlConstants.ATTRIBUTE_INSTANCE_NAME]
        pyutSDInstance.instanceLifeLineLength = SecureConversions.secureInteger(instanceElement[XmlConstants.ATTRIBUTE_LIFE_LINE_LENGTH])

        return pyutSDInstance

    def sdMessageToPyutSDMessage(self, oglSDMessageElement: Element) -> ConvolutedPyutSDMessageInformation:
        """
        TODO:  Need to fix how SD Messages are created
        Args:
            oglSDMessageElement:

        Returns:  Bogus data class
        """
        messageElement: Element = oglSDMessageElement.PyutSDMessage

        pyutSDMessage:  PyutSDMessage = PyutSDMessage()

        pyutSDMessage.id = int(messageElement['id'])
        pyutSDMessage.message = messageElement['message']
        pyutSDMessage.linkType = PyutLinkType.SD_MESSAGE

        srcID: int = int(messageElement[XmlConstants.ATTRIBUTE_SD_MESSAGE_SOURCE_ID])
        dstID: int = int(messageElement[XmlConstants.ATTRIBUTE_SD_MESSAGE_DESTINATION_ID])

        srcTime: int = int(messageElement[XmlConstants.ATTRIBUTE_SOURCE_TIME])
        dstTime: int = int(messageElement[XmlConstants.ATTRIBUTE_DESTINATION_TIME])

        pyutSDMessage.sourceY      = srcTime
        pyutSDMessage.destinationY = dstTime

        bogus: ConvolutedPyutSDMessageInformation = ConvolutedPyutSDMessageInformation()

        bogus.pyutSDMessage = pyutSDMessage
        bogus.sourceId      = srcID
        bogus.destinationId = dstID

        return bogus

    def _methodToPyutMethods(self, classElement: Element) -> PyutMethods:
        """
        The pyutClass may not have methods;
        Args:
            classElement:  The pyutClassElement

        Returns:  May return an empty list
        """
        untangledPyutMethods: PyutMethods = PyutMethods([])

        methodElements: Elements = cast(Elements, classElement.get_elements(XmlConstants.ELEMENT_MODEL_METHOD))

        for methodElement in methodElements:
            methodName: str            = methodElement['name']
            visibility: PyutVisibility = PyutVisibility.toEnum(methodElement[XmlConstants.ATTRIBUTE_VISIBILITY])
            self.logger.debug(f"{methodName=} - {visibility=}")

            pyutMethod: PyutMethod = PyutMethod(name=methodName, visibility=visibility)

            pyutMethod.modifiers = self._modifierToPyutMethodModifiers(methodElement=methodElement)

            returnAttribute = methodElement[XmlConstants.ATTRIBUTE_METHOD_RETURN_TYPE]
            pyutMethod.returnType = PyutType(returnAttribute)

            parameters = self._paramToPyutParameters(methodElement)
            pyutMethod.parameters = parameters
            pyutMethod.sourceCode = self._sourceCodeToPyutSourceCode(methodElement=methodElement)

            untangledPyutMethods.append(pyutMethod)

        return untangledPyutMethods

    def _fieldToPyutFields(self, classElement: Element) -> PyutFields:
        untangledPyutFields: PyutFields = PyutFields([])

        fieldElements: Elements = cast(Elements, classElement.get_elements(XmlConstants.ELEMENT_MODEL_FIELD))

        for fieldElement in fieldElements:
            visibility: PyutVisibility = PyutVisibility.toEnum(fieldElement['visibility'])
            fieldName    = fieldElement[XmlConstants.ATTRIBUTE_NAME]
            pyutType     = PyutType(fieldElement[XmlConstants.ATTRIBUTE_FIELD_TYPE])
            defaultValue = fieldElement[XmlConstants.ATTRIBUTE_DEFAULT_VALUE]

            pyutField: PyutField = PyutField(name=fieldName, visibility=visibility, type=pyutType, defaultValue=defaultValue)

            untangledPyutFields.append(pyutField)

        return untangledPyutFields

    def _modifierToPyutMethodModifiers(self, methodElement: Element) -> PyutModifiers:
        """
        Should be in this form:

            <Modifier name="Modifier1"/>
            <Modifier name="Modifier2"/>
            <Modifier name="Modifier3"/>
            <Modifier name="Modifier4"/>

        Args:
            methodElement:

        Returns:   A PyutModifiers object that may be empty.
        """

        modifierElements = methodElement.get_elements('Modifier')

        pyutModifiers: PyutModifiers = PyutModifiers([])
        if len(modifierElements) > 0:
            for modifierElement in modifierElements:
                modifierName:           str       = modifierElement['name']
                pyutModifier: PyutModifier = PyutModifier(name=modifierName)
                pyutModifiers.append(pyutModifier)

        return pyutModifiers

    def _paramToPyutParameters(self, methodElement: Element) -> PyutParameters:

        parameterElements = methodElement.get_elements(XmlConstants.ELEMENT_MODEL_PARAMETER)

        untangledPyutMethodParameters: PyutParameters = PyutParameters([])
        for parameterElement in parameterElements:
            name:           str = parameterElement['name']
            defaultValue:   str = parameterElement['defaultValue']
            parameterType:  PyutType = PyutType(parameterElement['type'])

            pyutParameter: PyutParameter = PyutParameter(name=name, type=parameterType, defaultValue=defaultValue)

            untangledPyutMethodParameters.append(pyutParameter)

        return untangledPyutMethodParameters

    def _sourceCodeToPyutSourceCode(self, methodElement: Element) -> SourceCode:

        sourceCodeElements = methodElement.get_elements(XmlConstants.ELEMENT_MODEL_SOURCE_CODE)
        codeElements = sourceCodeElements[0].get_elements(XmlConstants.ELEMENT_MODEL_CODE)
        sourceCode: SourceCode = SourceCode([])
        for codeElement in codeElements:
            self.logger.debug(f'{codeElement.cdata=}')
            codeLine: str = codeElement.cdata
            sourceCode.append(codeLine)
        return sourceCode

    def _interfaceMethodsToPyutMethods(self, interface: Element) -> PyutMethods:

        pyutMethods: PyutMethods = self._methodToPyutMethods(interface)

        return pyutMethods

    def _addPyutObjectAttributes(self, pyutElement: Element, pyutObject: PyutObject) -> PyutObject:
        """

        Args:
            pyutElement:    pyutElement XML with common keys
            pyutObject:     The PyutObject to update

        Returns:  The updated pyutObject as
        """

        pyutObject.id       = int(pyutElement[XmlConstants.ATTRIBUTE_ID])    # TODO revisit this when we start using UUIDs
        pyutObject.name     = pyutElement[XmlConstants.ATTRIBUTE_NAME]
        pyutObject.fileName = pyutElement[XmlConstants.ATTRIBUTE_FILENAME]

        if pyutObject.name is None:
            XmlToPyut.noteCounter += 1
            pyutObject.name = f'{XmlToPyut.NOTE_NAME}-{XmlToPyut.noteCounter}'
        return pyutObject

    def _securePyutDisplayMethods(self, displayStr: str) -> PyutDisplayMethods:

        if displayStr is not None:
            pyutDisplayMethods: PyutDisplayMethods = PyutDisplayMethods(displayStr)
        else:
            pyutDisplayMethods = PyutDisplayMethods.UNSPECIFIED

        return pyutDisplayMethods




from logging import Logger
from logging import getLogger

from pathlib import Path

from xml.etree.ElementTree import Element
from xml.etree.ElementTree import ElementTree
from xml.etree.ElementTree import SubElement

from xml.etree.ElementTree import indent as xmlIndent
from xml.etree.ElementTree import tostring as xmlToString
from xml.etree.ElementTree import fromstring as xmlFromString

from umlio.IOTypes import XML_VERSION
from umlio.IOTypes import UmlDocument

from umlio.serializer.UmlLinksToXml import UmlLinksToXml
from umlio.serializer.UmlNotesToXml import UmlNotesToXml
from umlio.serializer.UmlTextsToXml import UmlTextsToXml
from umlio.serializer.UmlClassToXml import UmlClassToXml
from umlio.serializer.UmlUseCasesToXml import UmlUseCasesToXml
from umlio.serializer.UmlLollipopsToXmlLollipops import UmlLollipopsToXmlLollipops

from umlio.XMLConstants import XmlConstants

INDENT_SPACES: str = '    '     # TODO: Make this configurable


class UmlShapesToXml:
    """
    The driver class to turn UML Shapes to XML
    """
    def __init__(self, projectFileName: Path, projectCodePath: Path):
        """
        Set up the XML Tree and the top level element

        Args:
            projectFileName:
            projectCodePath:
        """

        self.logger:           Logger = getLogger(__name__)
        self._projectCodePath: Path   = projectCodePath

        xmlProjectElement: Element = Element(XmlConstants.ELEMENT_UML_PROJECT)

        xmlProjectElement.set(XmlConstants.ATTRIBUTE_FILENAME, str(projectFileName))
        xmlProjectElement.set(XmlConstants.ATTRIBUTE_VERSION, XML_VERSION)
        xmlProjectElement.set(XmlConstants.ATTRIBUTE_CODE_PATH, str(projectCodePath))

        diagramTree: ElementTree = ElementTree(xmlProjectElement)

        xmlIndent(diagramTree, space='    ')

        self._diagramTree:       ElementTree = diagramTree
        self._xmlProjectElement: Element     = xmlProjectElement
        self._prettyPrint:       bool        = True

    @property
    def prettyPrint(self) -> bool:
        return self._prettyPrint

    @prettyPrint.setter
    def prettyPrint(self, prettyPrint: bool):
        self._prettyPrint = prettyPrint

    @property
    def xml(self) -> str:
        """

        Returns:  The serialized XML
        """
        if self.prettyPrint:
            return self._toPrettyString(self._xmlProjectElement)
        else:
            return self._toString(self._xmlProjectElement)

    def serialize(self, umlDiagram: UmlDocument):
        """
        Repeatedly call this method for each diagram in a UML project

        Args:
            umlDiagram:  The UML diagram to serialize

        """
        umlClassToXml:   UmlClassToXml    = UmlClassToXml()
        umlUseCaseToXml: UmlUseCasesToXml = UmlUseCasesToXml()
        umlNotesToXml:   UmlNotesToXml    = UmlNotesToXml()
        umlTextsToXml:   UmlTextsToXml    = UmlTextsToXml()
        umlLinksToXml:   UmlLinksToXml    = UmlLinksToXml()
        umlLollipopsToXml: UmlLollipopsToXmlLollipops = UmlLollipopsToXmlLollipops()

        documentElement: Element       = self._umlDocumentAttributesToXml(umlDiagram=umlDiagram)

        umlClassToXml.serialize(documentTop=documentElement, umlClasses=umlDiagram.umlClasses)
        umlUseCaseToXml.serialize(documentTop=documentElement, umlUseCases=umlDiagram.umlUseCases, umlActors=umlDiagram.umlActors)

        umlNotesToXml.serialize(documentTop=documentElement, umlNotes=umlDiagram.umlNotes)
        umlTextsToXml.serialize(documentTop=documentElement, umlTexts=umlDiagram.umlTexts)
        umlLinksToXml.serialize(documentTop=documentElement, umlLinks=umlDiagram.umlLinks)

        umlLollipopsToXml.serialize(documentTop=documentElement, umlLollipops=umlDiagram.umlLollipopInterfaces)

    def writeXml(self, fileName: Path):
        """
        Persist the XML

        Args:
            fileName:  The path object to the file
        """
        fileName.write_text(self.xml)

    def _umlDocumentAttributesToXml(self, umlDiagram: UmlDocument) -> Element:
        """
        Create a document sub element under the project element

        Args:
            umlDiagram:

        Returns:  The newly created document top sub element

        """

        attributes = {
            XmlConstants.ATTRIBUTE_DOCUMENT_TYPE:              umlDiagram.documentType.value,
            XmlConstants.ATTRIBUTE_TITLE:             umlDiagram.documentTitle,
            XmlConstants.ATTRIBUTE_SCROLL_POSITION_X: str(umlDiagram.scrollPositionX),
            XmlConstants.ATTRIBUTE_SCROLL_POSITION_Y: str(umlDiagram.scrollPositionY),
            XmlConstants.ATTRIBUTE_PIXELS_PER_UNIT_X: str(umlDiagram.pixelsPerUnitX),
            XmlConstants.ATTRIBUTE_PIXELS_PER_UNIT_Y: str(umlDiagram.pixelsPerUnitY),
        }
        documentTop: Element = SubElement(self._xmlProjectElement, XmlConstants.ELEMENT_UML_DIAGRAM, attrib=attributes)

        return documentTop

    def _toPrettyString(self, originalProjectElement: Element) -> str:
        """
        Create a copy of the input originalElement

        Convert to string, then parse again
        xmlToString() returns a binary, so we need to decode it to get a string

        Args:
            originalProjectElement:

        Returns:  An XML string
        """
        elementCopy: Element = xmlFromString(xmlToString(originalProjectElement))
        xmlIndent(elementCopy, space=INDENT_SPACES, level=0)

        return self._toString(elementCopy)

    def _toString(self, element: Element) -> str:
        return xmlToString(element, encoding='iso-8859-1', xml_declaration=True).decode('utf-8')




from logging import Logger
from logging import getLogger

from pathlib import Path

from untangle import Element
from untangle import parse

from codeallybasic.SecureConversions import SecureConversions

from umlshapes.ShapeTypes import LinkableUmlShapes
from umlshapes.ShapeTypes import linkableUmlShapesFactory

from umlio.IOTypes import UmlActors
from umlio.IOTypes import UmlClasses
from umlio.IOTypes import UmlDocument
from umlio.IOTypes import UmlDocumentTitle
from umlio.IOTypes import UmlDocumentType
from umlio.IOTypes import UmlLinks
from umlio.IOTypes import UmlLollipopInterfaces
from umlio.IOTypes import UmlNotes
from umlio.IOTypes import UmlTexts
from umlio.IOTypes import UmlProject
from umlio.IOTypes import UmlUseCases

from umlio.deserializer.XmlActorsToUmlActors import XmlActorsToUmlActors
from umlio.deserializer.XmlClassesToUmlClasses import XmlClassesToUmlClasses
from umlio.deserializer.XmlLinksToUmlLinks import XmlLinksToUmlLinks
from umlio.deserializer.XmlLollipopsToUmlLollipops import XmlLollipopsToUmlLollipops
from umlio.deserializer.XmlNotesToUmlNotes import XmlNotesToUmlNotes
from umlio.deserializer.XmlTextsToUmlTexts import XmlTextsToUmlTexts
from umlio.deserializer.XmlUseCasesToUmlUseCases import XmlUseCasesToUmlUseCases

from umlio.XMLConstants import XmlConstants


class XmlToUmlShapes:
    def __init__(self):
        self.logger: Logger = getLogger(__name__)

        self._umlProject:  UmlProject  = UmlProject()

    @property
    def umlProject(self) -> UmlProject:
        return self._umlProject

    def deserializeProjectFile(self, fileName: Path):
        pass

    def deserializeXmlFile(self, fileName: Path):
        """
        Untangle the input Xml string to UmlShapes
        Args:
            fileName:  The file name from which the XML came from
        """

        xmlString: str = fileName.read_text()
        self.deserializeXml(xmlString=xmlString, fileName=fileName)

    def deserializeXml(self, xmlString: str, fileName: Path):
        root:       Element = parse(xmlString)
        umlProject: Element = root.UmlProject

        self._umlProject.fileName = fileName
        self._umlProject.version  = umlProject[XmlConstants.ATTRIBUTE_VERSION]
        self._umlProject.codePath = umlProject[XmlConstants.ATTRIBUTE_CODE_PATH]

        for umlDiagramElement in umlProject.UMLDiagram:

            umlDocument: UmlDocument = UmlDocument(
                documentTitle=umlDiagramElement[XmlConstants.ATTRIBUTE_TITLE],
                scrollPositionX=SecureConversions.secureInteger(umlDiagramElement[XmlConstants.ATTRIBUTE_SCROLL_POSITION_X]),
                scrollPositionY=SecureConversions.secureInteger(umlDiagramElement[XmlConstants.ATTRIBUTE_SCROLL_POSITION_Y]),
                pixelsPerUnitX=SecureConversions.secureInteger(umlDiagramElement[XmlConstants.ATTRIBUTE_PIXELS_PER_UNIT_X]),
                pixelsPerUnitY=SecureConversions.secureInteger(umlDiagramElement[XmlConstants.ATTRIBUTE_PIXELS_PER_UNIT_Y])
            )
            umlDocument.documentTitle = UmlDocumentTitle(umlDiagramElement[XmlConstants.ATTRIBUTE_TITLE])

            if umlDiagramElement[XmlConstants.ATTRIBUTE_DOCUMENT_TYPE] == UmlDocumentType.CLASS_DOCUMENT.value:

                umlDocument.documentType = UmlDocumentType.CLASS_DOCUMENT
                umlDocument.umlClasses = self._deserializeUmlClassElements(umlDiagramElement=umlDiagramElement)
                umlDocument.umlNotes   = self._deserializeUmlNoteElements(umlDiagramElement=umlDiagramElement)
                umlDocument.umlTexts   = self._deserializeUmlTextElements(umlDiagramElement=umlDiagramElement)

                linkableUmlShapes: LinkableUmlShapes = self._buildLinkableUmlShapes(umlDocument=umlDocument)

                umlDocument.umlLinks              = self._deserializeUmlLinkElements(umlDiagramElement=umlDiagramElement, linkableUmlShapes=linkableUmlShapes)
                umlDocument.umlLollipopInterfaces = self._deserializeLollipopInterfaces(umlDiagramElement=umlDiagramElement, linkableUmlShapes=linkableUmlShapes)

            elif umlDiagramElement[XmlConstants.ATTRIBUTE_DOCUMENT_TYPE] == UmlDocumentType.USE_CASE_DOCUMENT.value:

                umlDocument.documentType = UmlDocumentType.USE_CASE_DOCUMENT
                umlDocument.umlNotes    = self._deserializeUmlNoteElements(umlDiagramElement=umlDiagramElement)
                umlDocument.umlActors   = self._deserializeUmlActorElements(umlDiagramElement=umlDiagramElement)
                umlDocument.umlUseCases = self._deserializeUmlUseCaseElements(umlDiagramElement=umlDiagramElement)

                linkableUmlShapes = self._buildLinkableUmlShapes(umlDocument=umlDocument)

                umlDocument.umlLinks = self._deserializeUmlLinkElements(umlDiagramElement=umlDiagramElement, linkableUmlShapes=linkableUmlShapes)

            elif umlDiagramElement[XmlConstants.ATTRIBUTE_DOCUMENT_TYPE] == UmlDocumentType.SEQUENCE_DOCUMENT.value:

                umlDocument.documentType = UmlDocumentType.SEQUENCE_DOCUMENT
                umlDocument.umlTexts   = self._deserializeUmlTextElements(umlDiagramElement=umlDiagramElement)
                umlDocument.umlNotes    = self._deserializeUmlNoteElements(umlDiagramElement=umlDiagramElement)

            else:
                assert False, 'Unknown diagram Type - Perhaps corrupted fle'

            self._umlProject.umlDocuments[umlDocument.documentTitle] = umlDocument

    def _deserializeUmlClassElements(self, umlDiagramElement: Element) -> UmlClasses:

        umlClassesDeSerializer: XmlClassesToUmlClasses = XmlClassesToUmlClasses()
        umlClasses:             UmlClasses             = umlClassesDeSerializer.deserialize(umlDiagramElement=umlDiagramElement)

        return umlClasses

    def _deserializeUmlTextElements(self, umlDiagramElement: Element) -> UmlTexts:
        """
        Yeah, yeah, I know bad English;

        Args:
            umlDiagramElement:  The diagram Element

        Returns:  deserialized UmlText objects if any exist, else an empty list
        """
        umlTextDeSerializer: XmlTextsToUmlTexts = XmlTextsToUmlTexts()
        umlTexts: UmlTexts = umlTextDeSerializer.deserialize(umlDiagramElement=umlDiagramElement)

        return umlTexts

    def _deserializeUmlNoteElements(self, umlDiagramElement: Element) -> UmlNotes:
        """

        Args:
            umlDiagramElement:  The diagram Element

        Returns:  deserialized UmlNote objects if any exist, else an empty list
        """
        xmlNotesToUmlNotes: XmlNotesToUmlNotes = XmlNotesToUmlNotes()
        umlNotes: UmlNotes = xmlNotesToUmlNotes.deserialize(umlDiagramElement=umlDiagramElement)

        return umlNotes

    def _deserializeUmlActorElements(self, umlDiagramElement: Element) -> UmlActors:
        """

        Args:
            umlDiagramElement:  The diagram Element

        Returns:  deserialized UmlActor objects if any exist, else an empty list
        """
        xmlActorsToUmlActors: XmlActorsToUmlActors = XmlActorsToUmlActors()

        umlActors: UmlActors = xmlActorsToUmlActors.deserialize(umlDiagramElement=umlDiagramElement)

        return umlActors

    def _deserializeUmlUseCaseElements(self, umlDiagramElement: Element) -> UmlUseCases:
        """

        Args:
            umlDiagramElement:  The diagram Element

        Returns:  deserialized UmlUseCase objects if any exist, else an empty list
        """
        xmlUseCasesToUmlUseCases: XmlUseCasesToUmlUseCases = XmlUseCasesToUmlUseCases()

        umlUseCases: UmlUseCases = xmlUseCasesToUmlUseCases.deserialize(umlDiagramElement=umlDiagramElement)

        return umlUseCases

    def _deserializeUmlLinkElements(self, umlDiagramElement: Element, linkableUmlShapes: LinkableUmlShapes) -> UmlLinks:

        xmlLinksToUmlLinks: XmlLinksToUmlLinks = XmlLinksToUmlLinks()

        umlLinks: UmlLinks = xmlLinksToUmlLinks.deserialize(umlDiagramElement=umlDiagramElement, linkableUmlShapes=linkableUmlShapes)

        return umlLinks

    def _deserializeLollipopInterfaces(self, umlDiagramElement: Element, linkableUmlShapes: LinkableUmlShapes) -> UmlLollipopInterfaces:

        xmlLollipopsToUmLollipops: XmlLollipopsToUmlLollipops = XmlLollipopsToUmlLollipops()

        lollipopInterfaces: UmlLollipopInterfaces = xmlLollipopsToUmLollipops.deserialize(umlDiagramElement=umlDiagramElement,
                                                                                          linkableUmlShapes=linkableUmlShapes
                                                                                          )
        return lollipopInterfaces

    def _buildLinkableUmlShapes(self, umlDocument: UmlDocument) -> LinkableUmlShapes:
        """

        Args:
            umlDocument:   The created document either Use case or class diagram

        Returns:  Linkable UML Shapes Dictionary
        """

        linkableUmlShapes: LinkableUmlShapes = linkableUmlShapesFactory()

        for umlClass in umlDocument.umlClasses:
            linkableUmlShapes[umlClass.pyutClass.id] = umlClass

        for umlNote in umlDocument.umlNotes:
            linkableUmlShapes[umlNote.pyutNote.id] = umlNote

        for umlUseCase in umlDocument.umlUseCases:
            linkableUmlShapes[umlUseCase.pyutUseCase.id] = umlUseCase

        for umlActor in umlDocument.umlActors:
            linkableUmlShapes[umlActor.pyutActor.id] = umlActor

        return linkableUmlShapes




from typing import Dict
from typing import List
from typing import NewType

from enum import Enum

from dataclasses import dataclass
from dataclasses import field

from pathlib import Path

from codeallybasic.SecureConversions import SecureConversions

from umlshapes.shapes.UmlActor import UmlActor
from umlshapes.shapes.UmlClass import UmlClass
from umlshapes.shapes.UmlNote import UmlNote
from umlshapes.shapes.UmlText import UmlText
from umlshapes.shapes.UmlUseCase import UmlUseCase
from umlshapes.links.UmlLollipopInterface import UmlLollipopInterface

from umlshapes.links.UmlLink import UmlLink
from umlshapes.types.UmlDimensions import UmlDimensions
from umlshapes.types.UmlPosition import UmlPosition
from untangle import Element

from umlio.XMLConstants import XmlConstants

XML_VERSION: str = '12.0'

PROJECT_SUFFIX: str = '.udt'        # UML Diagramming Tool
XML_SUFFIX:     str = '.xml'


UmlDocumentTitle = NewType('UmlDocumentTitle', str)
UmlClasses       = NewType('UmlClasses',      List[UmlClass])
UmlUseCases      = NewType('UmlUseCases',     List[UmlUseCase])
UmlActors        = NewType('UmlActors',       List[UmlActor])
UmlNotes         = NewType('UmlNotes',        List[UmlNote])
UmlTexts         = NewType('UmlTexts',        List[UmlText])
UmlLinks         = NewType('UmlLinks',        List[UmlLink])

UmlLollipopInterfaces = NewType('UmlLollipopInterfaces', List[UmlLollipopInterface])

ElementAttributes = NewType('ElementAttributes', Dict[str, str])


class UmlDocumentType(Enum):
    CLASS_DOCUMENT    = 'Class Document'
    USE_CASE_DOCUMENT = 'Use Case Document'
    SEQUENCE_DOCUMENT = 'Sequence Document'
    NOT_SET          = 'Not Set'


def umlClassesFactory() -> UmlClasses:
    """
    Factory method to create  the UmlClasses data structure;

    Returns:  A new data structure
    """
    return UmlClasses([])


def umlUseCasesFactory() -> UmlUseCases:
    return UmlUseCases([])


def umlActorsFactory() -> UmlActors:
    return UmlActors([])


def umlNotesFactory() -> UmlNotes:
    return UmlNotes([])


def umlTextsFactory() -> UmlTexts:
    return UmlTexts([])


def umlLinksFactory() -> UmlLinks:
    return UmlLinks([])


def umlLollipopInterfacesFactory() -> UmlLollipopInterfaces:
    return UmlLollipopInterfaces([])


@dataclass
class UmlDocument:
    documentType:    UmlDocumentType  = UmlDocumentType.NOT_SET
    documentTitle:   UmlDocumentTitle = UmlDocumentTitle('')
    scrollPositionX: int = 1
    scrollPositionY: int = 1
    pixelsPerUnitX:  int = 1
    pixelsPerUnitY:  int = 1
    umlClasses:      UmlClasses  = field(default_factory=umlClassesFactory)
    umlUseCases:     UmlUseCases = field(default_factory=umlUseCasesFactory)
    umlActors:       UmlActors   = field(default_factory=umlActorsFactory)
    umlNotes:        UmlNotes    = field(default_factory=umlNotesFactory)
    umlTexts:        UmlTexts    = field(default_factory=umlTextsFactory)
    umlLinks:        UmlLinks    = field(default_factory=umlLinksFactory)

    umlLollipopInterfaces: UmlLollipopInterfaces = field(default_factory=umlLollipopInterfacesFactory)


UmlDocuments = NewType('UmlDocuments', Dict[UmlDocumentTitle, UmlDocument])


def createUmlDocumentsFactory() -> UmlDocuments:
    return UmlDocuments({})


DEFAULT_PROJECT_TITLE: UmlDocumentTitle = UmlDocumentTitle('NewProject')
DEFAULT_PROJECT_PATH:  Path             = Path(f'{DEFAULT_PROJECT_TITLE}{PROJECT_SUFFIX}')

@dataclass
class ProjectInformation:
    fileName:    Path = Path('')
    version:     str  = XML_VERSION
    codePath:    Path = Path('')


@dataclass
class UmlProject(ProjectInformation):
    umlDocuments: UmlDocuments = field(default_factory=createUmlDocumentsFactory)

    @classmethod
    def emptyProject(cls) -> 'UmlProject':
        umlProject:  UmlProject  = UmlProject(DEFAULT_PROJECT_PATH)
        umlDocument: UmlDocument = UmlDocument(
            documentType=UmlDocumentType.CLASS_DOCUMENT,
            documentTitle=DEFAULT_PROJECT_TITLE
        )
        umlProject.umlDocuments[DEFAULT_PROJECT_TITLE] = umlDocument

        return umlProject


#
# Untangler helper types
#
Elements = NewType('Elements', List[Element])


@dataclass
class GraphicInformation:
    """
    Internal Class used to move information from a untangler element into Python
    """
    id:       str
    size:     UmlDimensions
    position: UmlPosition

    @classmethod
    def toGraphicInfo(cls, graphicElement: Element) -> 'GraphicInformation':

        graphicInformation: GraphicInformation = GraphicInformation(
            id=graphicElement[XmlConstants.ATTRIBUTE_ID],
            position=UmlPosition(
                x=int(graphicElement[XmlConstants.ATTRIBUTE_X]),
                y=int(graphicElement[XmlConstants.ATTRIBUTE_Y])
            ),
            size=UmlDimensions(
                width=SecureConversions.secureInteger(graphicElement[XmlConstants.ATTRIBUTE_WIDTH]),
                height=SecureConversions.secureInteger(graphicElement[XmlConstants.ATTRIBUTE_HEIGHT])
            )
        )

        return graphicInformation


@dataclass
class UmlLinkAttributes:

    #     f'        <UmlLink id="{UML_LINK_CANONICAL_MONIKER}" fromX="248" fromY="300" toX="190" toY="174" spline="False">\n'
    fromPosition: UmlPosition
    toPosition:   UmlPosition
    id:     str = 'NO ID'
    spline: bool = False

    @classmethod
    def fromGraphicLink(cls, linkElement: Element) -> 'UmlLinkAttributes':

        fromX: int = SecureConversions.secureInteger(linkElement[XmlConstants.ATTRIBUTE_LINK_FROM_X])
        fromY: int = SecureConversions.secureInteger(linkElement[XmlConstants.ATTRIBUTE_LINK_FROM_Y])
        toY:   int = int(linkElement[XmlConstants.ATTRIBUTE_LINK_TO_X])
        toX:   int = int(linkElement[XmlConstants.ATTRIBUTE_LINK_TO_Y])

        shapeId: str  = linkElement[XmlConstants.ATTRIBUTE_ID]
        spline:  bool = SecureConversions.secureBoolean(linkElement[XmlConstants.ATTRIBUTE_SPLINE])

        gla: UmlLinkAttributes = UmlLinkAttributes(
            fromPosition=UmlPosition(x=fromX, y=fromY),
            toPosition=UmlPosition(x=toX, y=toY),
            spline=spline,
            id=shapeId
        )

        return gla
